@@ -30,6 +30,8 @@
 #include "gdk-pixbuf-i18n.h"
 #include "gdk-pixbuf-animation.h"
 
+#include <glib/gstdio.h>
+
 typedef struct _GdkPixbufNonAnim GdkPixbufNonAnim;
 typedef struct _GdkPixbufNonAnimClass GdkPixbufNonAnimClass;
 
@@ -114,7 +116,7 @@ gdk_pixbuf_animation_get_type (void)
 
 /**
  * gdk_pixbuf_animation_new_from_file:
- * @filename: Name of file to load.
+ * @filename: Name of file to load, in the GLib file name encoding
  * @error: return location for error
  *
  * Creates a new animation by loading it from a file.  The file format is
@@ -137,12 +139,13 @@ gdk_pixbuf_animation_new_from_file (const char *filename,
 	guchar buffer [128];
 	GdkPixbufModule *image_module;
         gchar *display_name;
+        gboolean locked;
 
 	g_return_val_if_fail (filename != NULL, NULL);
         g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
         display_name = g_filename_display_name (filename);
-	f = fopen (filename, "rb");
+	f = g_fopen (filename, "rb");
 	if (!f) {
                 g_set_error (error,
                              G_FILE_ERROR,
@@ -181,7 +184,7 @@ gdk_pixbuf_animation_new_from_file (const char *filename,
                         return NULL;
                 }
 
-        _gdk_pixbuf_lock (image_module);
+        locked = _gdk_pixbuf_lock (image_module);
 
 	if (image_module->load_animation == NULL) {
 		GdkPixbuf *pixbuf;
@@ -245,10 +248,35 @@ gdk_pixbuf_animation_new_from_file (const char *filename,
         g_free (display_name);
 
  out_unlock:
-        _gdk_pixbuf_unlock (image_module);
+        if (locked)
+                _gdk_pixbuf_unlock (image_module);
 	return animation;
 }
 
+#ifdef G_OS_WIN32
+
+#undef gdk_pixbuf_animation_new_from_file
+
+GdkPixbufAnimation *
+gdk_pixbuf_animation_new_from_file (const char *filename,
+                                    GError    **error)
+{
+	gchar *utf8_filename =
+		g_locale_to_utf8 (filename, -1, NULL, NULL, error);
+	GdkPixbufAnimation *retval;
+
+	if (utf8_filename == NULL)
+		return NULL;
+
+	retval = gdk_pixbuf_animation_new_from_file_utf8 (utf8_filename, error);
+
+	g_free (utf8_filename);
+
+	return retval;
+}
+
+#endif
+
 /**
  * gdk_pixbuf_animation_ref:
  * @animation: An animation.
@@ -477,9 +505,9 @@ gdk_pixbuf_animation_iter_get_delay_time (GdkPixbufAnimationIter *iter)
  * 
  * Gets the current pixbuf which should be displayed; the pixbuf will
  * be the same size as the animation itself
- * (gdk_pixbuf_animation_get_width(),
- * gdk_pixbuf_animation_get_height()). This pixbuf should be displayed
- * for gdk_pixbuf_animation_get_delay_time() milliseconds.  The caller
+ * (gdk_pixbuf_animation_get_width(), gdk_pixbuf_animation_get_height()). 
+ * This pixbuf should be displayed for 
+ * gdk_pixbuf_animation_iter_get_delay_time() milliseconds.  The caller
  * of this function does not own a reference to the returned pixbuf;
  * the returned pixbuf will become invalid when the iterator advances
  * to the next frame, which may happen anytime you call
