@@ -25,6 +25,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include <string.h>
 
 /* #define OLE2_DND */
@@ -64,7 +65,7 @@ typedef enum {
 #ifdef OLE2_DND
 
 #define PRINT_GUID(guid) \
-  g_print ("guid = %.08x-%.04x-%.04x-%.02x%.02x-%.02x%.02x%.02x%.02x%.02x%.02x", \
+  g_print ("guid = %.08lx-%.04x-%.04x-%.02x%.02x-%.02x%.02x%.02x%.02x%.02x%.02x", \
 	   ((gulong *)  guid)[0], \
 	   ((gushort *) guid)[2], \
 	   ((gushort *) guid)[3], \
@@ -87,12 +88,14 @@ static int nformats;
  * this is used on both source and destination sides.
  */
 struct _GdkDragContextPrivateWin32 {
+#ifdef OLE2_DND
   gint    ref_count;
-
+#endif
   guint16 last_x;		/* Coordinates from last event */
   guint16 last_y;
-  HWND    dest_xid;
-  guint   drag_status;		/* Current status of drag */
+  HWND dest_xid;
+  guint drag_status : 4;	/* Current status of drag */
+  guint drop_failed : 1;	/* Whether the drop was unsuccessful */
 };
 
 #define GDK_DRAG_CONTEXT_PRIVATE_DATA(context) ((GdkDragContextPrivateWin32 *) GDK_DRAG_CONTEXT (context)->windowing_data)
@@ -140,7 +143,9 @@ gdk_drag_context_init (GdkDragContext *dragcontext)
   GdkDragContextPrivateWin32 *private = g_new0 (GdkDragContextPrivateWin32, 1);
 
   dragcontext->windowing_data = private;
+#ifdef OLE2_DND
   private->ref_count = 1;
+#endif
 
   contexts = g_list_prepend (contexts, dragcontext);
 }
@@ -160,7 +165,9 @@ gdk_drag_context_finalize (GObject *object)
 {
   GdkDragContext *context = GDK_DRAG_CONTEXT (object);
   GdkDragContextPrivateWin32 *private = GDK_DRAG_CONTEXT_PRIVATE_DATA (context);
-  
+ 
+  GDK_NOTE (DND, g_print ("gdk_drag_context_finalize\n"));
+ 
   g_list_free (context->targets);
 
   if (context->source_window)
@@ -173,6 +180,9 @@ gdk_drag_context_finalize (GObject *object)
   
   contexts = g_list_remove (contexts, context);
 
+  if (context == current_dest_drag)
+    current_dest_drag = NULL;
+
   g_free (private);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -191,8 +201,6 @@ gdk_drag_context_ref (GdkDragContext *context)
 {
   g_return_if_fail (GDK_IS_DRAG_CONTEXT (context));
 
-  GDK_NOTE (DND, g_print ("gdk_drag_context_ref: %p %d\n", context, G_OBJECT(context)->ref_count));
-
   g_object_ref (context);
 }
 
@@ -201,7 +209,6 @@ gdk_drag_context_unref (GdkDragContext *context)
 {
   g_return_if_fail (GDK_IS_DRAG_CONTEXT (context));
 
-  GDK_NOTE (DND, g_print ("gdk_drag_context_unref: %p %d\n", context, G_OBJECT(context)->ref_count));
   g_object_unref (context);
 }
 
@@ -668,7 +675,7 @@ ienumformatetc_next (LPENUMFORMATETC This,
   enum_formats *en = (enum_formats *) This;
   int i, n;
 
-  GDK_NOTE (DND, g_print ("ienumformatetc_next %p %d %d\n", This, en->ix, celt));
+  GDK_NOTE (DND, g_print ("ienumformatetc_next %p %d %ld\n", This, en->ix, celt));
 
   n = 0;
   for (i = 0; i < celt; i++)
@@ -694,7 +701,7 @@ ienumformatetc_skip (LPENUMFORMATETC This,
 {
   enum_formats *en = (enum_formats *) This;
 
-  GDK_NOTE (DND, g_print ("ienumformatetc_skip %p %d %d\n", This, en->ix, celt));
+  GDK_NOTE (DND, g_print ("ienumformatetc_skip %p %d %ld\n", This, en->ix, celt));
   en->ix += celt;
 
   return S_OK;
@@ -1171,6 +1178,7 @@ local_send_drop (GdkDragContext *context,
       GdkDragContextPrivateWin32 *private;
       private = GDK_DRAG_CONTEXT_PRIVATE_DATA (current_dest_drag);
 
+      /* Pass ownership of context to the event */
       tmp_event.dnd.type = GDK_DROP_START;
       tmp_event.dnd.window = current_dest_drag->dest_window;
       tmp_event.dnd.send_event = FALSE;
@@ -1180,6 +1188,8 @@ local_send_drop (GdkDragContext *context,
       tmp_event.dnd.x_root = private->last_x;
       tmp_event.dnd.y_root = private->last_y;
       
+      current_dest_drag = NULL;
+
       gdk_event_put (&tmp_event);
     }
 
@@ -1241,7 +1251,6 @@ gdk_drag_begin (GdkWindow *window,
   HRESULT hResult;
   DWORD dwEffect;
   HGLOBAL global;
-  FORMATETC format;
   STGMEDIUM medium;
 
   g_return_val_if_fail (window != NULL, NULL);
@@ -1282,7 +1291,7 @@ gdk_drag_begin (GdkWindow *window,
 			  (hResult == DRAGDROP_S_DROP ? "DRAGDROP_S_DROP" :
 			   (hResult == DRAGDROP_S_CANCEL ? "DRAGDROP_S_CANCEL" :
 			    (hResult == E_UNEXPECTED ? "E_UNEXPECTED" :
-			     g_strdup_printf ("%#.8x", hResult))))));
+			     g_strdup_printf ("%#.8lx", hResult))))));
 
   dobj->ido.lpVtbl->Release (&dobj->ido);
   ctx->ids.lpVtbl->Release (&ctx->ids);
@@ -1359,8 +1368,8 @@ gdk_drag_find_window_for_screen (GdkDragContext  *context,
 {
   find_window_enum_arg a;
 
-  a.x = x_root;
-  a.y = y_root;
+  a.x = x_root - _gdk_offset_x;
+  a.y = y_root - _gdk_offset_y;
   a.ignore = drag_window ? GDK_WINDOW_HWND (drag_window) : NULL;
   a.result = NULL;
 
@@ -1607,7 +1616,7 @@ gdk_drop_finish (GdkDragContext *context,
 	
   g_return_if_fail (context != NULL);
 
-  GDK_NOTE (DND, g_print ("gdk_drop_finish\n"));
+  GDK_NOTE (DND, g_print ("gdk_drop_finish"));
 
   private = GDK_DRAG_CONTEXT_PRIVATE_DATA (context);
 
@@ -1723,3 +1732,16 @@ gdk_drag_get_selection (GdkDragContext *context)
   else
     return GDK_NONE;
 }
+
+gboolean 
+gdk_drag_drop_succeeded (GdkDragContext *context)
+{
+  GdkDragContextPrivateWin32 *private;
+
+  g_return_val_if_fail (context != NULL, FALSE);
+
+  private = GDK_DRAG_CONTEXT_PRIVATE_DATA (context);
+
+  /* FIXME: Can we set drop_failed when the drop has failed? */
+  return !private->drop_failed;
+}
