@@ -61,12 +61,13 @@ enum {
 enum {
   PROP_0,
   PROP_LABEL,
+  PROP_IMAGE,
   PROP_RELIEF,
   PROP_USE_UNDERLINE,
   PROP_USE_STOCK,
   PROP_FOCUS_ON_CLICK,
   PROP_XALIGN,
-  PROP_YALIGN
+  PROP_YALIGN,
 };
 
 #define GTK_BUTTON_GET_PRIVATE(o)       (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_TYPE_BUTTON, GtkButtonPrivate))
@@ -76,10 +77,9 @@ struct _GtkButtonPrivate
 {
   gfloat       xalign;
   gfloat       yalign;
-  GtkSettings *settings;
-  guint        show_image_connection;
   GtkWidget   *image;
   guint        align_set : 1;
+  guint        image_is_stock : 1;
 };
 
 static void gtk_button_class_init     (GtkButtonClass   *klass);
@@ -293,6 +293,21 @@ gtk_button_class_init (GtkButtonClass *klass)
                                                       0.5,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
 
+  /**
+   * GtkButton::image:
+   * 
+   * The child widget to appear next to the button text.
+   * 
+   * Since: 2.6
+   */
+  g_object_class_install_property (gobject_class,
+                                   PROP_IMAGE,
+                                   g_param_spec_object ("image",
+                                                        P_("Image widget"),
+                                                        P_("Child widget to appear next to the button text"),
+                                                        GTK_TYPE_WIDGET,
+                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));
+
   button_signals[PRESSED] =
     g_signal_new ("pressed",
 		  G_OBJECT_CLASS_TYPE (object_class),
@@ -523,6 +538,9 @@ gtk_button_set_property (GObject         *object,
     case PROP_LABEL:
       gtk_button_set_label (button, g_value_get_string (value));
       break;
+    case PROP_IMAGE:
+      gtk_button_set_image (button, (GtkWidget *) g_value_get_object (value));
+      break;
     case PROP_RELIEF:
       gtk_button_set_relief (button, g_value_get_enum (value));
       break;
@@ -561,6 +579,9 @@ gtk_button_get_property (GObject         *object,
     case PROP_LABEL:
       g_value_set_string (value, button->label_text);
       break;
+    case PROP_IMAGE:
+      g_value_set_object (value, (GObject *)priv->image);
+      break;
     case PROP_RELIEF:
       g_value_set_enum (value, gtk_button_get_relief (button));
       break;
@@ -610,6 +631,8 @@ gtk_button_construct_child (GtkButton *button)
   GtkWidget *label;
   GtkWidget *hbox;
   GtkWidget *align;
+  GtkWidget *image = NULL;
+  gchar *label_text = NULL;
   
   if (!button->constructed)
     return;
@@ -619,19 +642,33 @@ gtk_button_construct_child (GtkButton *button)
 
   if (GTK_BIN (button)->child)
     {
+      if (priv->image && !priv->image_is_stock)
+	image = g_object_ref (priv->image);
+
       gtk_container_remove (GTK_CONTAINER (button),
   			    GTK_BIN (button)->child);
+  
       priv->image = NULL;
     }
   
   if (button->use_stock &&
       gtk_stock_lookup (button->label_text, &item))
     {
-      label = gtk_label_new_with_mnemonic (item.label);
+      if (!image)
+	image = g_object_ref (gtk_image_new_from_stock (button->label_text, GTK_ICON_SIZE_BUTTON));
 
+      label_text = item.label;
+    }
+  else
+    label_text = button->label_text;
+
+  if (image)
+    {
+      label = gtk_label_new_with_mnemonic (label_text);
       gtk_label_set_mnemonic_widget (GTK_LABEL (label), GTK_WIDGET (button));
       
-      priv->image = gtk_image_new_from_stock (button->label_text, GTK_ICON_SIZE_BUTTON);
+      priv->image = image;
+
       g_object_set (priv->image, 
 		    "visible", show_image (button),
 		    "no_show_all", TRUE,
@@ -650,10 +687,12 @@ gtk_button_construct_child (GtkButton *button)
       gtk_container_add (GTK_CONTAINER (align), hbox);
       gtk_widget_show_all (align);
 
+      g_object_unref (image);
+
       return;
     }
-
-  if (button->use_underline)
+  
+ if (button->use_underline)
     {
       label = gtk_label_new_with_mnemonic (button->label_text);
       gtk_label_set_mnemonic_widget (GTK_LABEL (label), GTK_WIDGET (button));
@@ -775,7 +814,6 @@ gtk_button_set_relief (GtkButton *button,
 GtkReliefStyle
 gtk_button_get_relief (GtkButton *button)
 {
-  g_return_val_if_fail (button != NULL, GTK_RELIEF_NORMAL);
   g_return_val_if_fail (GTK_IS_BUTTON (button), GTK_RELIEF_NORMAL);
 
   return button->relief;
@@ -1613,43 +1651,64 @@ show_image_change_notify (GtkButton *button)
   GtkButtonPrivate *priv = GTK_BUTTON_GET_PRIVATE (button);
 
   if (priv->image) 
-    g_object_set (priv->image, "visible", show_image (button), NULL);
+    {
+      if (show_image (button))
+	gtk_widget_show (priv->image);
+      else
+	gtk_widget_hide (priv->image);
+    }
 }
 
+static void
+traverse_container (GtkWidget *widget,
+		    gpointer   data)
+{
+  if (GTK_IS_BUTTON (widget))
+    show_image_change_notify (GTK_BUTTON (widget));
+  else if (GTK_IS_CONTAINER (widget))
+    gtk_container_forall (GTK_CONTAINER (widget), traverse_container, NULL);
+}
+
+static void
+gtk_button_setting_changed (GtkSettings *settings)
+{
+  GList *list, *l;
+
+  list = gtk_window_list_toplevels ();
+
+  for (l = list; l; l = l->next)
+    gtk_container_forall (GTK_CONTAINER (l->data), 
+			  traverse_container, NULL);
+
+  g_list_free (list);
+}
+
+
 static void
 gtk_button_screen_changed (GtkWidget *widget,
 			   GdkScreen *previous_screen)
 {
-  GtkButtonPrivate *priv = GTK_BUTTON_GET_PRIVATE (widget);
   GtkSettings *settings;
+  guint show_image_connection;
 
-  if (gtk_widget_has_screen (widget))
-    settings = gtk_widget_get_settings (widget);
-  else
-    settings = NULL;
-
-  if (settings == priv->settings)
+  if (!gtk_widget_has_screen (widget))
     return;
 
-  if (priv->settings)
-    {
-      g_signal_handler_disconnect (priv->settings, priv->show_image_connection);
-      g_object_unref (priv->settings);
-    }
+  settings = gtk_widget_get_settings (widget);
 
-  if (settings)
-    {
-      priv->show_image_connection =
-	g_signal_connect_swapped (settings,
-				  "notify::gtk-button-images",
-				  G_CALLBACK (show_image_change_notify),
-				  widget);
+  show_image_connection = 
+    GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (settings), 
+					 "gtk-button-connection"));
+  
+  if (show_image_connection)
+    return;
 
-      g_object_ref (settings);
-      priv->settings = settings;
-    }
-  else
-    priv->settings = NULL;
+  show_image_connection =
+    g_signal_connect (settings, "notify::gtk-button-images",
+		      G_CALLBACK (gtk_button_setting_changed), 0);
+  g_object_set_data (G_OBJECT (settings), 
+		     "gtk-button-connection",
+		     GUINT_TO_POINTER (show_image_connection));
 
   show_image_change_notify (GTK_BUTTON (widget));
 }
@@ -1679,3 +1738,52 @@ gtk_button_grab_notify (GtkWidget *widget,
       gtk_real_button_released (button);
     }
 }
+
+/**
+ * gtk_button_set_image:
+ * @button: a #GtkButton
+ * @image: a widget to set as the image for the button
+ *
+ * Set the image of @button to the given widget. Note that
+ * it depends on the show-button-images setting whether the
+ * image will be displayed or not.
+ *
+ * Since: 2.6
+ */ 
+void
+gtk_button_set_image (GtkButton *button,
+		      GtkWidget *image)
+{
+  GtkButtonPrivate *priv = GTK_BUTTON_GET_PRIVATE (button);
+
+  priv->image = image;
+  priv->image_is_stock = FALSE;
+
+  gtk_button_construct_child (button);
+
+  g_object_notify (G_OBJECT (button), "image");
+}
+
+/**
+ * gtk_button_get_image:
+ * @button: a #GtkButton
+ *
+ * Gets the widget that is currenty set as the image of @button.
+ * This may have been explicitly set by gtk_button_set_image()
+ * or constructed by gtk_button_new_from_stock().
+ *
+ * Since: 2.6
+ */
+GtkWidget *
+gtk_button_get_image (GtkButton *button)
+{
+  GtkButtonPrivate *priv;
+
+  g_return_val_if_fail (GTK_IS_BUTTON (button), NULL);
+
+  priv = GTK_BUTTON_GET_PRIVATE (button);
+  
+  return priv->image;
+}
+  
+  
