@@ -17,8 +17,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <stdlib.h>
+#include "gtkalias.h"
 #include "gtkcellrendererpixbuf.h"
+#include "gtkiconfactory.h"
 #include "gtkintl.h"
 
 static void gtk_cell_renderer_pixbuf_get_property  (GObject                    *object,
@@ -164,10 +167,11 @@ gtk_cell_renderer_pixbuf_class_init (GtkCellRendererPixbufClass *class)
 
   g_object_class_install_property (object_class,
 				   PROP_STOCK_SIZE,
-				   g_param_spec_enum ("stock_size",
+				   g_param_spec_uint ("stock_size",
 						      P_("Size"),
-						      P_("The size of the rendered icon"),
-						      GTK_TYPE_ICON_SIZE,
+						      P_("The GtkIconSize value that specifies the size of the rendered icon"),
+						      0,
+						      G_MAXUINT,
 						      GTK_ICON_SIZE_MENU,
 						      G_PARAM_READWRITE));
 
@@ -190,7 +194,7 @@ gtk_cell_renderer_pixbuf_finalize (GObject *object)
 
   priv = GTK_CELL_RENDERER_PIXBUF_GET_PRIVATE (object);
 
-  if (cellpixbuf->pixbuf && priv->stock_id)
+  if (cellpixbuf->pixbuf)
     g_object_unref (cellpixbuf->pixbuf);
 
   if (priv->stock_id)
@@ -231,7 +235,7 @@ gtk_cell_renderer_pixbuf_get_property (GObject        *object,
       g_value_set_string (value, priv->stock_id);
       break;
     case PROP_STOCK_SIZE:
-      g_value_set_enum (value, priv->stock_size);
+      g_value_set_uint (value, priv->stock_size);
       break;
     case PROP_STOCK_DETAIL:
       g_value_set_string (value, priv->stock_detail);
@@ -294,7 +298,7 @@ gtk_cell_renderer_pixbuf_set_property (GObject      *object,
       priv->stock_id = g_strdup (g_value_get_string (value));
       break;
     case PROP_STOCK_SIZE:
-      priv->stock_size = g_value_get_enum (value);
+      priv->stock_size = g_value_get_uint (value);
       break;
     case PROP_STOCK_DETAIL:
       if (priv->stock_detail)
@@ -429,6 +433,7 @@ gtk_cell_renderer_pixbuf_render (GtkCellRenderer      *cell,
   GtkCellRendererPixbuf *cellpixbuf = (GtkCellRendererPixbuf *) cell;
   GtkCellRendererPixbufPrivate *priv;
   GdkPixbuf *pixbuf;
+  GdkPixbuf *invisible = NULL;
   GdkRectangle pix_rect;
   GdkRectangle draw_rect;
   gboolean stock_pixbuf = FALSE;
@@ -465,6 +470,33 @@ gtk_cell_renderer_pixbuf_render (GtkCellRenderer      *cell,
   pix_rect.width  -= cell->xpad * 2;
   pix_rect.height -= cell->ypad * 2;
 
+  if (GTK_WIDGET_STATE (widget) == GTK_STATE_INSENSITIVE || !cell->sensitive)
+    {
+      GtkIconSource *source;
+      
+      source = gtk_icon_source_new ();
+      gtk_icon_source_set_pixbuf (source, pixbuf);
+      /* The size here is arbitrary; since size isn't
+       * wildcarded in the souce, it isn't supposed to be
+       * scaled by the engine function
+       */
+      gtk_icon_source_set_size (source, GTK_ICON_SIZE_SMALL_TOOLBAR);
+      gtk_icon_source_set_size_wildcarded (source, FALSE);
+      
+     invisible = gtk_style_render_icon (widget->style,
+					source,
+					gtk_widget_get_direction (widget),
+					GTK_STATE_INSENSITIVE,
+					/* arbitrary */
+					(GtkIconSize)-1,
+					widget,
+					"gtkcellrendererpixbuf");
+     
+     gtk_icon_source_free (source);
+     
+     pixbuf = invisible;
+    }
+
   if (gdk_rectangle_intersect (cell_area, &pix_rect, &draw_rect) &&
       gdk_rectangle_intersect (expose_area, &draw_rect, &draw_rect))
     gdk_draw_pixbuf (window,
@@ -479,4 +511,7 @@ gtk_cell_renderer_pixbuf_render (GtkCellRenderer      *cell,
 		     draw_rect.height,
 		     GDK_RGB_DITHER_NORMAL,
 		     0, 0);
+  
+  if (invisible)
+    g_object_unref (invisible);
 }
