@@ -79,7 +79,7 @@ GQuark     gtk_file_system_error_quark      (void);
  */
 #define GTK_TYPE_FILE_INFO (gtk_file_info_get_type ())
 
-GType       gtk_file_info_get_type (void);
+GType       gtk_file_info_get_type (void) G_GNUC_CONST; 
 
 GtkFileInfo *gtk_file_info_new  (void);
 GtkFileInfo *gtk_file_info_copy (GtkFileInfo *info);
@@ -201,7 +201,7 @@ struct _GtkFileSystemIface
   void (*bookmarks_changed) (GtkFileSystem *file_system);
 };
 
-GType             gtk_file_system_get_type       (void);
+GType             gtk_file_system_get_type       (void) G_GNUC_CONST;
 
 GSList *          gtk_file_system_list_volumes   (GtkFileSystem     *file_system);
 
@@ -307,9 +307,13 @@ struct _GtkFileFolderIface
 			 GSList        *paths);
   void (*files_removed) (GtkFileFolder *monitor,
 			 GSList        *paths);
+
+  /* Method / signal */
+  gboolean (*is_finished_loading) (GtkFileFolder *folder);
+  void     (*finished_loading)    (GtkFileFolder *folder);
 };
 
-GType        gtk_file_folder_get_type      (void);
+GType        gtk_file_folder_get_type      (void) G_GNUC_CONST;
 gboolean     gtk_file_folder_list_children (GtkFileFolder      *folder,
 					    GSList            **children,
 					    GError            **error);
@@ -317,13 +321,13 @@ GtkFileInfo *gtk_file_folder_get_info      (GtkFileFolder      *folder,
 					    const GtkFilePath  *path,
 					    GError            **error);
 
+gboolean     gtk_file_folder_is_finished_loading (GtkFileFolder *folder);
+
 
 /* GtkFilePath */
 #define GTK_TYPE_FILE_PATH             (gtk_file_path_get_type ())
-#define GTK_FILE_PATH(obj)             (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_FILE_PATH, GtkFilePath))
-#define GTK_IS_FILE_PATH(obj)          (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_FILE_PATH))
 
-GType   gtk_file_path_get_type (void);
+GType   gtk_file_path_get_type (void) G_GNUC_CONST;
 #ifdef __GNUC__
 #define gtk_file_path_new_dup(str) \
  ({ const gchar *__s = (str); (GtkFilePath *)g_strdup(__s); })
