@@ -27,6 +27,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <glib.h>
+#include <glib/gstdio.h>
 #include "gtkalias.h"
 
 #ifdef G_OS_WIN32
@@ -36,7 +37,7 @@
 #endif /* G_OS_WIN32 */
 
 #include "gtkicontheme.h"
-#include "gtkiconthemeparser.h"
+#include "gtkiconcache.h"
 #include "gtkintl.h"
 #include "gtksettings.h"
 #include "gtkprivate.h"
@@ -61,7 +62,8 @@ typedef enum
   ICON_SUFFIX_NONE = 0,
   ICON_SUFFIX_XPM = 1 << 0,
   ICON_SUFFIX_SVG = 1 << 1,
-  ICON_SUFFIX_PNG = 1 << 2
+  ICON_SUFFIX_PNG = 1 << 2,
+  HAS_ICON_FILE = 1 << 3
 } IconSuffix;
 
 
@@ -81,7 +83,8 @@ struct _GtkIconThemePrivate
    */
   GList *themes;
   GHashTable *unthemed_icons;
-
+  GList *unthemed_icons_caches;
+  
   /* Note: The keys of this hashtable are owned by the
    * themedir and unthemed hashtables.
    */
@@ -101,6 +104,12 @@ struct _GtkIconInfo
   /* Information about the source
    */
   gchar *filename;
+#ifdef G_OS_WIN32
+  /* System codepage version of filename, for DLL ABI backward
+   * compatibility functions.
+   */
+  gchar *cp_filename;
+#endif
   GdkPixbuf *builtin_pixbuf;
 
   GtkIconData *data;
@@ -132,6 +141,11 @@ typedef struct
   char *comment;
   char *example;
 
+  /* Icon caches, per theme directory, key is NULL if
+   * no cache exists for that directory
+   */
+  GHashTable *icon_caches;
+  
   /* In search order */
   GList *dirs;
 } IconTheme;
@@ -158,6 +172,9 @@ typedef struct
   int threshold;
 
   char *dir;
+  char *subdir;
+  
+  GtkIconCache *cache;
   
   GHashTable *icons;
   GHashTable *icon_data;
@@ -197,13 +214,21 @@ static void         theme_list_icons  (IconTheme        *theme,
 				       GQuark            context);
 static void         theme_subdir_load (GtkIconTheme     *icon_theme,
 				       IconTheme        *theme,
-				       GtkIconThemeFile *theme_file,
+				       GKeyFile         *theme_file,
 				       char             *subdir);
 static void         do_theme_change   (GtkIconTheme     *icon_theme);
 
 static void  blow_themes               (GtkIconTheme    *icon_themes);
 
 static void  icon_data_free            (GtkIconData          *icon_data);
+static void load_icon_data (IconThemeDir *dir,
+			    const char   *path,
+			    const char   *name);
+
+static IconSuffix theme_dir_get_icon_suffix (IconThemeDir *dir,
+					     const gchar  *icon_name,
+					     gboolean     *has_icon_file);
+
 
 static GtkIconInfo *icon_info_new             (void);
 static GtkIconInfo *icon_info_new_builtin     (BuiltinIcon *icon);
@@ -495,12 +520,16 @@ gtk_icon_theme_set_screen (GtkIconTheme *icon_theme,
  * with GdkPixbuf.
  */
 static gboolean
-pixbuf_supports_svg ()
+pixbuf_supports_svg (void)
 {
   GSList *formats = gdk_pixbuf_get_formats ();
   GSList *tmp_list;
-  gboolean found_svg = FALSE;
+  static gboolean found_svg = FALSE;
+  static gboolean value_known = FALSE;
 
+  if (value_known)
+    return found_svg;
+  
   for (tmp_list = formats; tmp_list && !found_svg; tmp_list = tmp_list->next)
     {
       gchar **mime_types = gdk_pixbuf_format_get_mime_types (tmp_list->data);
@@ -516,7 +545,8 @@ pixbuf_supports_svg ()
     }
 
   g_slist_free (formats);
-    
+  value_known = TRUE;
+  
   return found_svg;
 }
 
@@ -553,7 +583,8 @@ gtk_icon_theme_init (GtkIconTheme *icon_theme)
   priv->themes_valid = FALSE;
   priv->themes = NULL;
   priv->unthemed_icons = NULL;
-
+  priv->unthemed_icons_caches = NULL;
+  
   priv->pixbuf_supports_svg = pixbuf_supports_svg ();
 }
 
@@ -569,6 +600,8 @@ do_theme_change (GtkIconTheme *icon_theme)
 {
   GtkIconThemePrivate *priv = icon_theme->priv;
   
+  GTK_NOTE (ICONTHEME, 
+	    g_print ("change to icon theme \"%s\"\n", priv->current_theme));
   blow_themes (icon_theme);
   g_signal_emit (icon_theme, signal_changed, 0);
   
@@ -579,6 +612,16 @@ do_theme_change (GtkIconTheme *icon_theme)
     }
 }
 
+static void 
+free_cache (gpointer data, 
+	    gpointer user_data)
+{
+  GtkIconCache *cache = (GtkIconCache *)data;
+
+  if (cache)
+    _gtk_icon_cache_unref (cache);
+}
+
 static void
 blow_themes (GtkIconTheme *icon_theme)
 {
@@ -592,9 +635,13 @@ blow_themes (GtkIconTheme *icon_theme)
       g_list_foreach (priv->dir_mtimes, (GFunc)free_dir_mtime, NULL);
       g_list_free (priv->dir_mtimes);
       g_hash_table_destroy (priv->unthemed_icons);
+      if (priv->unthemed_icons_caches)
+	g_list_foreach (priv->unthemed_icons_caches, free_cache, NULL);
+      g_list_free (priv->unthemed_icons_caches);
     }
   priv->themes = NULL;
   priv->unthemed_icons = NULL;
+  priv->unthemed_icons_caches = NULL;
   priv->dir_mtimes = NULL;
   priv->all_icons = NULL;
   priv->themes_valid = FALSE;
@@ -715,7 +762,8 @@ gtk_icon_theme_get_search_path (GtkIconTheme      *icon_theme,
  * @icon_theme: a #GtkIconTheme
  * @path: directory name to append to the icon path
  * 
- * Appends a directory to the search path. See gtk_icon_theme_set_search_path(). 
+ * Appends a directory to the search path. 
+ * See gtk_icon_theme_set_search_path(). 
  *
  * Since: 2.4
  **/
@@ -742,7 +790,8 @@ gtk_icon_theme_append_search_path (GtkIconTheme *icon_theme,
  * @icon_theme: a #GtkIconTheme
  * @path: directory name to prepend to the icon path
  * 
- * Prepends a directory to the search path. See gtk_icon_theme_set_search_path().
+ * Prepends a directory to the search path. 
+ * See gtk_icon_theme_set_search_path().
  *
  * Since: 2.4
  **/
@@ -822,15 +871,13 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
   GtkIconThemePrivate *priv;
   IconTheme *theme;
   char *path;
-  char *contents;
-  char *directories;
-  char *inherits;
-  GtkIconThemeFile *theme_file;
+  GKeyFile *theme_file;
+  GError *error = NULL;
   IconThemeDirMtime *dir_mtime;
   struct stat stat_buf;
   
   priv = icon_theme->priv;
-  
+
   for (l = priv->themes; l != NULL; l = l->next)
     {
       theme = l->data;
@@ -845,29 +892,35 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
 			       NULL);
       dir_mtime = g_new (IconThemeDirMtime, 1);
       dir_mtime->dir = path;
-      if (stat (path, &stat_buf) == 0 && S_ISDIR (stat_buf.st_mode))
+      if (g_stat (path, &stat_buf) == 0 && S_ISDIR (stat_buf.st_mode))
 	dir_mtime->mtime = stat_buf.st_mtime;
       else
 	dir_mtime->mtime = 0;
 
       priv->dir_mtimes = g_list_prepend (priv->dir_mtimes, dir_mtime);
     }
+  priv->dir_mtimes = g_list_reverse (priv->dir_mtimes);
 
   theme_file = NULL;
-  for (i = 0; i < priv->search_path_len; i++)
+  for (i = 0; i < priv->search_path_len && !theme_file; i++)
     {
       path = g_build_filename (priv->search_path[i],
 			       theme_name,
 			       "index.theme",
 			       NULL);
-      if (g_file_test (path, G_FILE_TEST_IS_REGULAR)) {
-	if (g_file_get_contents (path, &contents, NULL, NULL)) {
-	  theme_file = _gtk_icon_theme_file_new_from_string (contents, NULL);
-	  g_free (contents);
-	  g_free (path);
-	  break;
+      if (g_file_test (path, G_FILE_TEST_IS_REGULAR)) 
+	{
+	  theme_file = g_key_file_new ();
+	  g_key_file_set_list_separator (theme_file, ',');
+	  g_key_file_load_from_file (theme_file, path, 0, &error);
+	  if (error)
+	    {
+	      g_key_file_free (theme_file);
+	      theme_file = NULL;
+	      g_error_free (error);
+	      error = NULL;
+	    }
 	}
-      }
       g_free (path);
     }
 
@@ -875,70 +928,62 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
     return;
   
   theme = g_new (IconTheme, 1);
-  if (!_gtk_icon_theme_file_get_locale_string (theme_file,
-					       "Icon Theme",
-					       "Name",
-					       &theme->display_name))
+  theme->display_name = 
+    g_key_file_get_locale_string (theme_file, "Icon Theme", "Name", NULL, NULL);
+  if (!theme->display_name)
     {
       g_warning ("Theme file for %s has no name\n", theme_name);
       g_free (theme);
-      _gtk_icon_theme_file_free (theme_file);
+      g_key_file_free (theme_file);
       return;
     }
 
-  if (!_gtk_icon_theme_file_get_string (theme_file,
-					"Icon Theme",
-					"Directories",
-					&directories))
+  dirs = g_key_file_get_string_list (theme_file, "Icon Theme", "Directories", NULL, NULL);
+  if (!dirs)
     {
       g_warning ("Theme file for %s has no directories\n", theme_name);
       g_free (theme->display_name);
       g_free (theme);
-      _gtk_icon_theme_file_free (theme_file);
+      g_key_file_free (theme_file);
       return;
     }
   
   theme->name = g_strdup (theme_name);
-  _gtk_icon_theme_file_get_locale_string (theme_file,
-					  "Icon Theme",
-					  "Comment",
-					  &theme->comment);
-  _gtk_icon_theme_file_get_string (theme_file,
-				   "Icon Theme",
-				   "Example",
-				   &theme->example);
-  
-  dirs = g_strsplit (directories, ",", 0);
+  theme->comment = 
+    g_key_file_get_locale_string (theme_file, 
+				  "Icon Theme", "Comment",
+				  NULL, NULL);
+  theme->example = 
+    g_key_file_get_string (theme_file, 
+			   "Icon Theme", "Example",
+			   NULL);
 
+  theme->icon_caches = NULL;
   theme->dirs = NULL;
   for (i = 0; dirs[i] != NULL; i++)
-      theme_subdir_load (icon_theme, theme, theme_file, dirs[i]);
+    theme_subdir_load (icon_theme, theme, theme_file, dirs[i]);
   
   g_strfreev (dirs);
   
   theme->dirs = g_list_reverse (theme->dirs);
 
-  g_free (directories);
-
   /* Prepend the finished theme */
   priv->themes = g_list_prepend (priv->themes, theme);
 
-  if (_gtk_icon_theme_file_get_string (theme_file,
+  themes = g_key_file_get_string_list (theme_file,
 				       "Icon Theme",
 				       "Inherits",
-				       &inherits))
+				       NULL,
+				       NULL);
+  if (themes)
     {
-      themes = g_strsplit (inherits, ",", 0);
-
       for (i = 0; themes[i] != NULL; i++)
 	insert_theme (icon_theme, themes[i]);
       
       g_strfreev (themes);
-
-      g_free (inherits);
     }
 
-  _gtk_icon_theme_file_free (theme_file);
+  g_key_file_free (theme_file);
 }
 
 static void
@@ -973,13 +1018,25 @@ load_themes (GtkIconTheme *icon_theme)
   /* Always look in the "default" icon theme */
   insert_theme (icon_theme, DEFAULT_THEME_NAME);
   priv->themes = g_list_reverse (priv->themes);
-  
+
+
   priv->unthemed_icons = g_hash_table_new_full (g_str_hash, g_str_equal,
 						g_free, (GDestroyNotify)free_unthemed_icon);
 
   for (base = 0; base < icon_theme->priv->search_path_len; base++)
     {
+      GtkIconCache *cache;
       dir = icon_theme->priv->search_path[base];
+
+      cache = _gtk_icon_cache_new_for_path (dir);
+
+      if (cache != NULL)
+	{
+	  priv->unthemed_icons_caches = g_list_prepend (priv->unthemed_icons_caches, cache);
+
+	  continue;
+	}
+	   
       gdir = g_dir_open (dir, 0, NULL);
 
       if (gdir == NULL)
@@ -1036,7 +1093,7 @@ load_themes (GtkIconTheme *icon_theme)
 		    unthemed_icon->svg_filename = abs_file;
 		  else
 		    unthemed_icon->no_svg_filename = abs_file;
-		  
+
 		  g_hash_table_insert (priv->unthemed_icons,
 				       base_name,
 				       unthemed_icon);
@@ -1062,7 +1119,7 @@ ensure_valid_themes (GtkIconTheme *icon_theme)
   
   if (priv->themes_valid)
     {
-      g_get_current_time(&tv);
+      g_get_current_time (&tv);
 
       if (ABS (tv.tv_sec - priv->last_stat_time) > 5)
 	gtk_icon_theme_rescan_if_needed (icon_theme);
@@ -1109,7 +1166,7 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
   g_return_val_if_fail (icon_name != NULL, NULL);
   g_return_val_if_fail ((flags & GTK_ICON_LOOKUP_NO_SVG) == 0 ||
 			(flags & GTK_ICON_LOOKUP_FORCE_SVG) == 0, NULL);
-  
+
   priv = icon_theme->priv;
 
   if (flags & GTK_ICON_LOOKUP_NO_SVG)
@@ -1127,12 +1184,12 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
   l = priv->themes;
   while (l != NULL)
     {
-      IconTheme *icon_theme = l->data;
+      IconTheme *theme = l->data;
       
-      if (strcmp (icon_theme->name, DEFAULT_THEME_NAME) == 0)
+      if (strcmp (theme->name, DEFAULT_THEME_NAME) == 0)
 	found_default = TRUE;
       
-      icon_info = theme_lookup_icon (icon_theme, icon_name, size, allow_svg, use_builtin);
+      icon_info = theme_lookup_icon (theme, icon_name, size, allow_svg, use_builtin);
       if (icon_info)
 	goto out;
       
@@ -1164,6 +1221,10 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
 	icon_info->filename = g_strdup (unthemed_icon->svg_filename);
       else if (unthemed_icon->no_svg_filename)
 	icon_info->filename = g_strdup (unthemed_icon->no_svg_filename);
+#ifdef G_OS_WIN32
+      icon_info->cp_filename = g_locale_from_utf8 (icon_info->filename,
+						   -1, NULL, NULL, NULL);
+#endif
 
       icon_info->dir_type = ICON_THEME_DIR_UNTHEMED;
     }
@@ -1268,6 +1329,27 @@ gtk_icon_theme_load_icon (GtkIconTheme         *icon_theme,
   return pixbuf;
 }
 
+typedef struct 
+{
+  const gchar *icon_name;
+  gboolean found;
+} CacheSearch;
+
+static void
+cache_has_icon (gpointer  key,
+		gpointer  value,
+		gpointer  user_data)
+{
+  GtkIconCache *cache = (GtkIconCache *)value;
+  CacheSearch *search = (CacheSearch *)user_data;
+
+  if (!cache || search->found)
+    return;
+
+  if (_gtk_icon_cache_has_icon (cache, search->icon_name))
+    search->found = TRUE;  
+}
+
 /**
  * gtk_icon_theme_has_icon:
  * @icon_theme: a #GtkIconTheme
@@ -1286,6 +1368,8 @@ gtk_icon_theme_has_icon (GtkIconTheme *icon_theme,
 			 const char   *icon_name)
 {
   GtkIconThemePrivate *priv;
+  GList *l;
+  CacheSearch search;
 
   g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), FALSE);
   
@@ -1293,6 +1377,27 @@ gtk_icon_theme_has_icon (GtkIconTheme *icon_theme,
   
   ensure_valid_themes (icon_theme);
 
+  search.icon_name = icon_name;
+  search.found = FALSE;
+
+  for (l = priv->themes; l; l = l->next)
+    {
+      IconTheme *theme = (IconTheme *)l->data;
+
+      g_hash_table_foreach (theme->icon_caches, cache_has_icon, &search);
+
+      if (search.found)
+	return TRUE;
+    }
+
+  for (l = priv->unthemed_icons_caches; l; l = l->next)
+    {
+      GtkIconCache *cache = (GtkIconCache *)l->data;
+
+      if (_gtk_icon_cache_has_icon (cache, icon_name))
+	return TRUE;
+    }
+
   if (g_hash_table_lookup_extended (priv->all_icons,
 				    icon_name, NULL, NULL))
     return TRUE;
@@ -1328,7 +1433,8 @@ add_size (gpointer  key,
  * is zero-terminated.
  * 
  * Return value: An newly allocated array describing the sizes at
- * which the icon is available.
+ * which the icon is available. The array should be freed with g_free()
+ * when it is no longer needed.
  *
  * Since: 2.6
  **/
@@ -1343,7 +1449,7 @@ gtk_icon_theme_get_icon_sizes (GtkIconTheme *icon_theme,
   
   GtkIconThemePrivate *priv;
 
-  g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), FALSE);
+  g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), NULL);
   
   priv = icon_theme->priv;
 
@@ -1357,8 +1463,8 @@ gtk_icon_theme_get_icon_sizes (GtkIconTheme *icon_theme,
       for (d = theme->dirs; d; d = d->next)
 	{
 	  IconThemeDir *dir = d->data;
-	  
-	  suffix = GPOINTER_TO_UINT (g_hash_table_lookup (dir->icons, icon_name));
+
+	  suffix = theme_dir_get_icon_suffix (dir, icon_name, NULL);	  
 	  if (suffix != ICON_SUFFIX_NONE)
 	    {
 	      if (suffix == ICON_SUFFIX_SVG)
@@ -1453,7 +1559,7 @@ gtk_icon_theme_list_icons (GtkIconTheme *icon_theme,
 			  add_key_to_hash,
 			  icons);
 
-  list = 0;
+  list = NULL;
   
   g_hash_table_foreach (icons,
 			add_key_to_list,
@@ -1534,7 +1640,7 @@ gtk_icon_theme_rescan_if_needed (GtkIconTheme *icon_theme)
     {
       dir_mtime = d->data;
 
-      stat_res = stat (dir_mtime->dir, &stat_buf);
+      stat_res = g_stat (dir_mtime->dir, &stat_buf);
 
       /* dir mtime didn't change */
       if (stat_res == 0 && 
@@ -1566,16 +1672,25 @@ theme_destroy (IconTheme *theme)
 
   g_list_foreach (theme->dirs, (GFunc)theme_dir_destroy, NULL);
   g_list_free (theme->dirs);
+  
+  if (theme->icon_caches)
+    g_hash_table_destroy (theme->icon_caches);
+
   g_free (theme);
 }
 
 static void
 theme_dir_destroy (IconThemeDir *dir)
 {
-  g_hash_table_destroy (dir->icons);
+  if (dir->cache)
+      _gtk_icon_cache_unref (dir->cache);
+  else
+    g_hash_table_destroy (dir->icons);
+  
   if (dir->icon_data)
     g_hash_table_destroy (dir->icon_data);
   g_free (dir->dir);
+  g_free (dir->subdir);
   g_free (dir);
 }
 
@@ -1663,6 +1778,31 @@ best_suffix (IconSuffix suffix,
     return ICON_SUFFIX_NONE;
 }
 
+
+static IconSuffix
+theme_dir_get_icon_suffix (IconThemeDir *dir,
+			   const gchar  *icon_name,
+			   gboolean     *has_icon_file)
+{
+  IconSuffix suffix;
+
+  if (dir->cache)
+    {
+      suffix = (IconSuffix)_gtk_icon_cache_get_icon_flags (dir->cache,
+							   icon_name,
+							   dir->subdir);
+
+      if (has_icon_file)
+	*has_icon_file = suffix & HAS_ICON_FILE;
+
+      suffix = suffix & ~HAS_ICON_FILE;
+    }
+  else
+      suffix = GPOINTER_TO_UINT (g_hash_table_lookup (dir->icons, icon_name));
+
+  return suffix;
+}
+
 static GtkIconInfo *
 theme_lookup_icon (IconTheme          *theme,
 		   const char         *icon_name,
@@ -1700,10 +1840,8 @@ theme_lookup_icon (IconTheme          *theme,
     {
       dir = l->data;
 
-      suffix = GPOINTER_TO_UINT (g_hash_table_lookup (dir->icons, icon_name));
-      
-      if (suffix != ICON_SUFFIX_NONE &&
-	  (allow_svg || suffix != ICON_SUFFIX_SVG))
+      suffix = theme_dir_get_icon_suffix (dir, icon_name, NULL);
+      if (best_suffix (suffix, allow_svg) != ICON_SUFFIX_NONE)
 	{
 	  difference = theme_dir_size_difference (dir, size, &smaller);
 
@@ -1744,15 +1882,37 @@ theme_lookup_icon (IconTheme          *theme,
   if (min_dir)
     {
       GtkIconInfo *icon_info = icon_info_new ();
+      gboolean has_icon_file;
       
-      suffix = GPOINTER_TO_UINT (g_hash_table_lookup (min_dir->icons, icon_name));
+      suffix = theme_dir_get_icon_suffix (min_dir, icon_name, &has_icon_file);
       suffix = best_suffix (suffix, allow_svg);
       g_assert (suffix != ICON_SUFFIX_NONE);
       
       file = g_strconcat (icon_name, string_from_suffix (suffix), NULL);
       icon_info->filename = g_build_filename (min_dir->dir, file, NULL);
       g_free (file);
+#ifdef G_OS_WIN32
+      icon_info->cp_filename = g_locale_from_utf8 (icon_info->filename,
+						   -1, NULL, NULL, NULL);
+#endif
+      if (min_dir->cache && has_icon_file)
+	{
+	  gchar *icon_file_name, *icon_file_path;
 
+	  icon_file_name = g_strconcat (icon_name, ".icon", NULL);
+	  icon_file_path = g_build_filename (min_dir->dir, icon_file_name, NULL);
+
+	  if (g_file_test (icon_file_path, G_FILE_TEST_IS_REGULAR))
+	    {
+	      if (min_dir->icon_data == NULL)
+		min_dir->icon_data = g_hash_table_new_full (g_str_hash, g_str_equal,
+							    g_free, (GDestroyNotify)icon_data_free);
+	      load_icon_data (min_dir, icon_file_path, icon_file_name);
+	    }
+	  g_free (icon_file_name);
+	  g_free (icon_file_path);
+	}
+      
       if (min_dir->icon_data != NULL)
 	icon_info->data = g_hash_table_lookup (min_dir->icon_data, icon_name);
 
@@ -1779,10 +1939,21 @@ theme_list_icons (IconTheme *theme, GHashTable *icons,
 
       if (context == dir->context ||
 	  context == 0)
-	g_hash_table_foreach (dir->icons,
-			      add_key_to_hash,
-			      icons);
-
+	{
+	  if (dir->cache)
+	    {
+	      _gtk_icon_cache_add_icons (dir->cache,
+					 dir->subdir,
+					 icons);
+					 
+	    }
+	  else
+	    {
+	      g_hash_table_foreach (dir->icons,
+				    add_key_to_hash,
+				    icons);
+	    }
+	}
       l = l->next;
     }
 }
@@ -1790,94 +1961,84 @@ theme_list_icons (IconTheme *theme, GHashTable *icons,
 static void
 load_icon_data (IconThemeDir *dir, const char *path, const char *name)
 {
-  GtkIconThemeFile *icon_file;
+  GKeyFile *icon_file;
   char *base_name;
   char **split;
-  char *contents;
+  gsize length;
   char *dot;
   char *str;
   char *split_point;
   int i;
+  gint *ivalues;
+  GError *error = NULL;
   
   GtkIconData *data;
 
-  if (g_file_get_contents (path, &contents, NULL, NULL))
+  icon_file = g_key_file_new ();
+  g_key_file_set_list_separator (icon_file, ',');
+  g_key_file_load_from_file (icon_file, path, 0, &error);
+  if (error)
     {
-      icon_file = _gtk_icon_theme_file_new_from_string (contents, NULL);
+      g_error_free (error);
+      return;
+    }
+  else
+    {
+      base_name = g_strdup (name);
+      dot = strrchr (base_name, '.');
+      *dot = 0;
       
-      if (icon_file)
+      data = g_new0 (GtkIconData, 1);
+      g_hash_table_replace (dir->icon_data, base_name, data);
+      
+      ivalues = g_key_file_get_integer_list (icon_file, 
+					     "Icon Data", "EmbeddedTextRectangle",
+					      &length, NULL);
+      if (ivalues)
 	{
-	  base_name = g_strdup (name);
-	  dot = strrchr (base_name, '.');
-	  *dot = 0;
+	  if (length == 4)
+	    {
+	      data->has_embedded_rect = TRUE;
+	      data->x0 = ivalues[0];
+	      data->y0 = ivalues[1];
+	      data->x1 = ivalues[2];
+	      data->y1 = ivalues[3];
+	    }
 	  
-	  data = g_new0 (GtkIconData, 1);
-	  g_hash_table_replace (dir->icon_data, base_name, data);
+	  g_free (ivalues);
+	}
+      
+      str = g_key_file_get_string (icon_file, "Icon Data", "AttachPoints", NULL);
+      if (str)
+	{
+	  split = g_strsplit (str, "|", -1);
 	  
-	  if (_gtk_icon_theme_file_get_string (icon_file, "Icon Data",
-					       "EmbeddedTextRectangle",
-					       &str))
-	    {
-	      split = g_strsplit (str, ",", 4);
-	      
-	      i = 0;
-	      while (split[i] != NULL)
-		i++;
+	  data->n_attach_points = g_strv_length (split);
+	  data->attach_points = g_malloc (sizeof (GdkPoint) * data->n_attach_points);
 
-	      if (i==4)
-		{
-		  data->has_embedded_rect = TRUE;
-		  data->x0 = atoi (split[0]);
-		  data->y0 = atoi (split[1]);
-		  data->x1 = atoi (split[2]);
-		  data->y1 = atoi (split[3]);
-		}
-
-	      g_strfreev (split);
-	      g_free (str);
-	    }
-
-
-	  if (_gtk_icon_theme_file_get_string (icon_file, "Icon Data",
-					       "AttachPoints",
-					       &str))
+	  i = 0;
+	  while (split[i] != NULL && i < data->n_attach_points)
 	    {
-	      split = g_strsplit (str, "|", -1);
-	      
-	      i = 0;
-	      while (split[i] != NULL)
-		i++;
-
-	      data->n_attach_points = i;
-	      data->attach_points = g_malloc (sizeof (GdkPoint) * i);
-
-	      i = 0;
-	      while (split[i] != NULL && i < data->n_attach_points)
+	      split_point = strchr (split[i], ',');
+	      if (split_point)
 		{
-		  split_point = strchr (split[i], ',');
-		  if (split_point)
-		    {
-		      *split_point = 0;
-		      split_point++;
-		      data->attach_points[i].x = atoi (split[i]);
-		      data->attach_points[i].y = atoi (split_point);
-		    }
-		  i++;
+		  *split_point = 0;
+		  split_point++;
+		  data->attach_points[i].x = atoi (split[i]);
+		  data->attach_points[i].y = atoi (split_point);
 		}
-	      
-	      g_strfreev (split);
-	      g_free (str);
+	      i++;
 	    }
 	  
-	  _gtk_icon_theme_file_get_locale_string (icon_file, "Icon Data",
-						  "DisplayName",
-						  &data->display_name);
-	  
-	  _gtk_icon_theme_file_free (icon_file);
+	  g_strfreev (split);
+	  g_free (str);
 	}
-      g_free (contents);
+      
+      data->display_name = g_key_file_get_locale_string (icon_file, 
+							 "Icon Data", "DisplayName",
+							 NULL, NULL);
+      g_key_file_free (icon_file);
     }
-  
 }
 
 static void
@@ -1889,7 +2050,9 @@ scan_directory (GtkIconThemePrivate *icon_theme,
   char *base_name, *dot;
   char *path;
   IconSuffix suffix, hash_suffix;
-  
+
+  GTK_NOTE (ICONTHEME, 
+	    g_print ("scanning directory %s\n", full_dir));
   dir->icons = g_hash_table_new_full (g_str_hash, g_str_equal,
 				      g_free, NULL);
   
@@ -1933,11 +2096,11 @@ scan_directory (GtkIconThemePrivate *icon_theme,
 
 static void
 theme_subdir_load (GtkIconTheme *icon_theme,
-		   IconTheme *theme,
-		   GtkIconThemeFile *theme_file,
-		   char *subdir)
+		   IconTheme    *theme,
+		   GKeyFile     *theme_file,
+		   char         *subdir)
 {
-  int base;
+  GList *d;
   char *type_string;
   IconThemeDir *dir;
   IconThemeDirType type;
@@ -1948,18 +2111,22 @@ theme_subdir_load (GtkIconTheme *icon_theme,
   int max_size;
   int threshold;
   char *full_dir;
+  GError *error = NULL;
+  GtkIconCache *cache;
+  IconThemeDirMtime *dir_mtime;
 
-  if (!_gtk_icon_theme_file_get_integer (theme_file,
-					 subdir,
-					 "Size",
-					 &size))
+  size = g_key_file_get_integer (theme_file, subdir, "Size", &error);
+  if (error)
     {
-      g_warning ("Theme directory %s of theme %s has no size field\n", subdir, theme->name);
+      g_error_free (error);
+      g_warning ("Theme directory %s of theme %s has no size field\n", 
+		 subdir, theme->name);
       return;
     }
   
   type = ICON_THEME_DIR_THRESHOLD;
-  if (_gtk_icon_theme_file_get_string (theme_file, subdir, "Type", &type_string))
+  type_string = g_key_file_get_string (theme_file, subdir, "Type", NULL);
+  if (type_string)
     {
       if (strcmp (type_string, "Fixed") == 0)
 	type = ICON_THEME_DIR_FIXED;
@@ -1972,37 +2139,64 @@ theme_subdir_load (GtkIconTheme *icon_theme,
     }
   
   context = 0;
-  if (_gtk_icon_theme_file_get_string (theme_file, subdir, "Context", &context_string))
+  context_string = g_key_file_get_string (theme_file, subdir, "Context", NULL);
+  if (context_string)
     {
       context = g_quark_from_string (context_string);
       g_free (context_string);
     }
 
-  if (!_gtk_icon_theme_file_get_integer (theme_file,
-					 subdir,
-					 "MaxSize",
-				     &max_size))
-    max_size = size;
-  
-  if (!_gtk_icon_theme_file_get_integer (theme_file,
-					 subdir,
-					 "MinSize",
-					 &min_size))
-    min_size = size;
-  
-  if (!_gtk_icon_theme_file_get_integer (theme_file,
-					 subdir,
-					 "Threshold",
-					 &threshold))
-    threshold = 2;
-
-  for (base = 0; base < icon_theme->priv->search_path_len; base++)
+  max_size = g_key_file_get_integer (theme_file, subdir, "MaxSize", &error);
+  if (error)
     {
-      full_dir = g_build_filename (icon_theme->priv->search_path[base],
-				   theme->name,
-				   subdir,
-				   NULL);
-      if (g_file_test (full_dir, G_FILE_TEST_IS_DIR))
+      max_size = size;
+
+      g_error_free (error);
+      error = NULL;
+    }
+
+  min_size = g_key_file_get_integer (theme_file, subdir, "MinSize", &error);
+  if (error)
+    {
+      min_size = size;
+
+      g_error_free (error);
+      error = NULL;
+    }
+  
+  threshold = g_key_file_get_integer (theme_file, subdir, "Threshold", &error);
+  if (error)
+    {
+      threshold = 2;
+
+      g_error_free (error);
+      error = NULL;
+    }
+
+  for (d = icon_theme->priv->dir_mtimes; d; d = d->next)
+    {
+      dir_mtime = (IconThemeDirMtime *)d->data;
+
+      if (dir_mtime->mtime == 0)
+	continue; /* directory doesn't exist */
+
+       full_dir = g_build_filename (dir_mtime->dir, subdir, NULL);
+
+      /* First, see if we have a cache for the directory */
+      if (!theme->icon_caches)
+	theme->icon_caches = g_hash_table_new_full (g_str_hash, g_str_equal,
+						    g_free, (GDestroyNotify)free_cache);
+						   
+      if (!g_hash_table_lookup_extended (theme->icon_caches, dir_mtime->dir, 
+					 NULL, (gpointer)&cache))
+	{
+	  /* This will return NULL if the cache doesn't exist or is outdated */
+	  cache = _gtk_icon_cache_new_for_path (dir_mtime->dir);
+
+	  g_hash_table_insert (theme->icon_caches, g_strdup (dir_mtime->dir), cache);
+	}
+
+      if (cache != NULL || g_file_test (full_dir, G_FILE_TEST_IS_DIR))
 	{
 	  dir = g_new (IconThemeDir, 1);
 	  dir->type = type;
@@ -2013,8 +2207,14 @@ theme_subdir_load (GtkIconTheme *icon_theme,
 	  dir->threshold = threshold;
 	  dir->dir = full_dir;
 	  dir->icon_data = NULL;
-	  
-	  scan_directory (icon_theme->priv, dir, full_dir);
+	  dir->subdir = g_strdup (subdir);
+	  if (cache != NULL)
+	    dir->cache = _gtk_icon_cache_ref (cache);
+	  else
+	    {
+	      dir->cache = NULL;
+	      scan_directory (icon_theme->priv, dir, full_dir);
+	    }
 
 	  theme->dirs = g_list_prepend (theme->dirs, dir);
 	}
@@ -2096,6 +2296,10 @@ gtk_icon_info_copy (GtkIconInfo *icon_info)
     copy->load_error = g_error_copy (copy->load_error);
   if (copy->filename)
     copy->filename = g_strdup (copy->filename);
+#ifdef G_OS_WIN32
+  if (copy->cp_filename)
+    copy->cp_filename = g_strdup (copy->cp_filename);
+#endif
 
   return copy;
 }
@@ -2115,6 +2319,10 @@ gtk_icon_info_free (GtkIconInfo *icon_info)
 
   if (icon_info->filename)
     g_free (icon_info->filename);
+#ifdef G_OS_WIN32
+  if (icon_info->cp_filename)
+    g_free (icon_info->cp_filename);
+#endif
   if (icon_info->builtin_pixbuf)
     g_object_unref (icon_info->builtin_pixbuf);
   if (icon_info->pixbuf)
@@ -2321,12 +2529,12 @@ icon_info_ensure_scale_and_pixbuf (GtkIconInfo *icon_info,
     {
       gint image_size = MAX (image_width, image_height);
       if (image_size > 0)
-	icon_info->scale = icon_info->desired_size / image_size;
+	icon_info->scale = icon_info->desired_size / (gdouble)image_size;
       else
 	icon_info->scale = 1.0;
       
       if (icon_info->dir_type == ICON_THEME_DIR_UNTHEMED)
-	icon_info->scale = MAX (icon_info->scale, 1.0);
+	icon_info->scale = MIN (icon_info->scale, 1.0);
     }
 
   /* We don't short-circuit out here for scale_only, since, now
@@ -2362,7 +2570,8 @@ icon_info_ensure_scale_and_pixbuf (GtkIconInfo *icon_info,
  * pixbuf may not be exactly this size; an icon theme may have icons
  * that differ slightly from their nominal sizes, and in addition GTK+
  * will avoid scaling icons that it considers sufficiently close to the
- * requested size. (This maintains sharpness.)
+ * requested size or for which the source image would have to be scaled
+ * up too far. (This maintains sharpness.) 
  * 
  * Return value: the rendered icon; this may be a newly created icon
  *  or a new reference to an internal icon, so you must not modify
@@ -2606,7 +2815,7 @@ gtk_icon_info_get_display_name  (GtkIconInfo *icon_info)
  * that the icon is generally available.
  *
  * This function will generally be used with pixbufs loaded
- * via gdk_pixbuf_new_from_inline ().
+ * via gdk_pixbuf_new_from_inline().
  *
  * Since: 2.4
  **/
@@ -2713,3 +2922,94 @@ find_builtin_icon (const gchar *icon_name,
 
   return min_icon;
 }
+
+#ifdef G_OS_WIN32
+
+/* DLL ABI stability backward compatibility versions */
+
+#undef gtk_icon_theme_set_search_path
+
+void
+gtk_icon_theme_set_search_path (GtkIconTheme *icon_theme,
+				const gchar  *path[],
+				gint          n_elements)
+{
+  const gchar **utf8_path;
+  gint i;
+
+  utf8_path = g_new (const gchar *, n_elements);
+
+  for (i = 0; i < n_elements; i++)
+    utf8_path[i] = g_locale_to_utf8 (path[i], -1, NULL, NULL, NULL);
+
+  gtk_icon_theme_set_search_path_utf8 (icon_theme, utf8_path, n_elements);
+
+  for (i = 0; i < n_elements; i++)
+    g_free ((gchar *) utf8_path[i]);
+
+  g_free (utf8_path);
+}
+
+#undef gtk_icon_theme_get_search_path
+
+void
+gtk_icon_theme_get_search_path (GtkIconTheme      *icon_theme,
+				gchar            **path[],
+				gint              *n_elements)
+{
+  gint i, n;
+
+  gtk_icon_theme_get_search_path_utf8 (icon_theme, path, &n);
+
+  if (n_elements)
+    *n_elements = n;
+
+  if (path)
+    {
+      for (i = 0; i < n; i++)
+	{
+	  gchar *tem = (*path)[i];
+
+	  (*path)[i] = g_locale_from_utf8 ((*path)[i], -1, NULL, NULL, NULL);
+	  g_free (tem);
+	}
+    }
+}
+
+#undef gtk_icon_theme_append_search_path
+
+void
+gtk_icon_theme_append_search_path (GtkIconTheme *icon_theme,
+				   const gchar  *path)
+{
+  gchar *utf8_path = g_locale_from_utf8 (path, -1, NULL, NULL, NULL);
+
+  gtk_icon_theme_append_search_path_utf8 (icon_theme, utf8_path);
+
+  g_free (utf8_path);
+}
+
+#undef gtk_icon_theme_prepend_search_path
+
+void
+gtk_icon_theme_prepend_search_path (GtkIconTheme *icon_theme,
+				    const gchar  *path)
+{
+  gchar *utf8_path = g_locale_from_utf8 (path, -1, NULL, NULL, NULL);
+
+  gtk_icon_theme_prepend_search_path_utf8 (icon_theme, utf8_path);
+
+  g_free (utf8_path);
+}
+
+#undef gtk_icon_info_get_filename
+
+G_CONST_RETURN gchar *
+gtk_icon_info_get_filename (GtkIconInfo *icon_info)
+{
+  g_return_val_if_fail (icon_info != NULL, NULL);
+
+  return icon_info->cp_filename;
+}
+
+#endif
